/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "sound.h"
#if (USESOUNDSDL) || (USESOUNDSDLMIXER)

#  include <time.h>
#  include "jomar3D.h"
#  include "utils.h"

static int loadname (char name[MAXCMD]);
static void initializemusic (void);
static int switchmusic (void *dummy);

#  define MAXMUSIC 100

#  if (USESOUNDSDL)
#    include "core.h"
#    include "interface.h"
typedef struct
{
  SDL_AudioSpec spec;
  Uint8 *sound;                 /* Pointer to wave data */
  Uint32 soundlen;              /* Length of wave data */
  int soundpos;                 /* Current play position */
  char playable;                /* whether load is OK and play possible */
} sound;

int mixwall, mixtrycatch, mixpickup, mixwon, mixchou;

static int cylpos[NCYL];        // remember the position of each cyl's music

SDL_AudioSpec *obtained;
#  endif

#  if (USESOUNDSDLMIXER)
typedef struct
{
  Mix_Chunk *sound;
  char playable;                //whether load is OK and play possible
} sound;
#  endif

static int nmusic = 0;
static char files[MAXMUSIC][MAXCMD];
static sound wave, wall, trycatch, pickup, won, walk, menu, chou;

/* say if the music is playable 
int have_sound (int sn)
{
#define chk_case(c,v) case c: return (v.playable)?1:0; 

  switch (sn) {
    chk_case (walk); 
    chk_case (walk); 
    chk_case (walk); 
    chk_case (walk); 
    chk_case (walk); 
    
  }
}*/


/* load all musics */
static int
loadallmusics (void *dummy)
{
  (void) dummy;

#  define LOAD(name)                           \
      if (loadname(SOUNDDIR"/"#name".gwav")) {  \
         name = wave;                           \
         wave.sound = NULL;                     \
         name.playable = 1;                        \
      }                                         \
      else {                                    \
         name.playable = 0;                        \
      }

/* load the game's sounds */
  LOAD (wall)
    LOAD (trycatch)
    LOAD (pickup) LOAD (won) LOAD (walk) LOAD (menu) LOAD (chou)
#  undef LOAD
#  if (USESOUNDSDLMIXER)
    if (walk.playable)
    Mix_VolumeChunk (walk.sound, MIX_MAX_VOLUME * 0.2);
#  endif

/* select, load and launch the first music */
  switchmusic (NULL);

  return 1;
}

/* change the background music */
void
changemusic (void)
{
#  if (USESOUNDSDL)
  wave.soundpos = wave.soundlen;
#  elif (USESOUNDSDLMIXER)
  if (playsound)
    Mix_FadeOutChannel (0, 1000);
#  endif
}

void
launchmusic (void)
{
  nosound = !playsound;
#  if (USESOUNDSDL)
  mixwall = mixtrycatch = mixpickup = mixwon = mixchou = 0;
#  endif
  if (!nosound)
    {
      initializemusic ();
      if (!nosound)
        SDL_CreateThread (loadallmusics, NULL); // nosound could have changed
    }
}

/* mix a sound given its index number */
void
mixsound (int soundnb)
{
  if (!playsound)
    return;

#  if (DEBUG)
  printf ("mixing sound number %d\n", soundnb);
#  endif

#  if (USESOUNDSDL)
#    define MIX(name)                           \
       if (name.playable) mix##name=1;
#  elif (USESOUNDSDLMIXER)
#    define MIX(name)                           \
       if (name.playable) if (Mix_Playing(soundnb) == 0) Mix_PlayChannel(soundnb, name.sound, 0);
#  endif

  switch (soundnb)
    {
    case WALL:
      MIX (wall);
      break;
    case TRYCATCH:
      MIX (trycatch);
      break;
    case PICKUP:
      MIX (pickup);
      break;
    case WON:
      MIX (won);
      break;
#  if (USESOUNDSDLMIXER)
    case WALK:
    case BOOST:
      MIX (walk);
      break;
#  endif
    case CHOU:
      MIX (chou);
      break;
    case SMENU1:
    case SMENU2:
    case SMENU3:
    case SMENU4:
    case SMENU5:
    case SMENU6:
    case SMENU7:
    case SMENU8:
    case SMENU9:
    case SMENU10:
#  if (USESOUNDSDL)
      if (menu.playable)
        cylpos[soundnb - SMENU1] = 1;
#  elif (USESOUNDSDLMIXER)
      MIX (menu);
#  endif
      break;
    default:
      printf ("No sound %d to mix\n", soundnb);
    }
#  undef MIX
}





#  if (USESOUNDSDL)
/* mix a sample to the sound output */
static void
playmixsample (sound * base, Uint8 * stream, unsigned int len, float volume)
{
  if (base->soundlen - base->soundpos <= len)
    {
      SDL_MixAudio (stream, base->sound + base->soundpos,
                    base->soundlen - base->soundpos,
                    SDL_MIX_MAXVOLUME * volume);
      base->soundpos = 0;
    }
  else
    {
      SDL_MixAudio (stream, base->sound + base->soundpos, len,
                    SDL_MIX_MAXVOLUME * volume);
      base->soundpos += len;
    }
}

/* fill the sound buffer (automatically called by SDL every time needed */
static void
fillerup (void *unused, Uint8 * stream, int len)
{
  (void) unused;
  Uint8 *waveptr;
  int waveleft, i;

  if (!playsound)
    return;

  /* Set up the pointers */
  waveptr = wave.sound + wave.soundpos;
  waveleft = wave.soundlen - wave.soundpos;

#    if (DEBUG)
  printf ("wave left, sample %d %d \n", waveleft, len);
#    endif
  /* Go! */
  if (waveleft <= len)
    {
      if (waveleft != 0)
        {
          SDL_MixAudio (stream, waveptr, waveleft, SDL_MIX_MAXVOLUME);
          wave.soundpos = wave.soundlen;
        }
      else
        {
          SDL_FreeWAV (wave.sound);
          wave.soundpos = 0;
#    if (DEBUG)
          printf ("music terminated\n");
#    endif
          wave.spec.callback = NULL;
          switchmusic (NULL);
          //    SDL_CreateThread (switchmusic, NULL);   cannot load on another thread: segmentation fault !
        }
      return;
    }

  if (mixwall)
    {
      playmixsample (&wall, stream, len, 1.0);
      if (wall.soundpos == 0.)
        mixwall = 0;
    }
  if (mixtrycatch)
    {
      playmixsample (&trycatch, stream, len, 1.0);
      if (trycatch.soundpos == 0.)
        mixtrycatch = 0;
    }
  if (mixpickup)
    {
      playmixsample (&pickup, stream, len, 1.0);
      if (pickup.soundpos == 0.)
        mixpickup = 0;
    }
  if (mixwon)
    {
      playmixsample (&won, stream, len, 1.0);
      if (won.soundpos == 0.)
        mixwon = 0;
    }
  if (going)
    {
      if (walk.playable)
        {
          playmixsample (&walk, stream, len, 0.2);
        }
    }
  if (mixchou)
    {
      playmixsample (&chou, stream, len, 1.0);
      if (chou.soundpos == 0.)
        mixchou = 0;
    }
  if (currentmenu != NOMENU)
    {
      for (i = 0; i < NCYL; i++)
        {
          if (cylpos[i])
            {
              menu.soundpos = cylpos[i];
              if (menu.soundpos == 1)
                menu.soundpos = 0;
              playmixsample (&menu, stream, len, 1.0 - (float) (i) * 0.15);
              cylpos[i] = menu.soundpos;
            }
        }
    }

  SDL_MixAudio (stream, waveptr, len, SDL_MIX_MAXVOLUME);
  wave.soundpos += len;
}

/* load a sound file */
int
loadname (char name[MAXCMD])
{
#    if (DEBUG)
  printf ("loading %s\n", name);
#    endif

  /* Load the wave file into memory */
  if (SDL_LoadWAV (name, &wave.spec, &wave.sound, &wave.soundlen) == NULL)
    {
      fprintf (stderr, "Couldn't load %s: %s\n", name, SDL_GetError ());
      return 0;
    }

  wave.spec.callback = fillerup;

  SDL_AudioCVT wav_cvt;
  int ret;

  /* conversion only if no mixer (done on-the-fly) */
  /* Build AudioCVT */
  ret = SDL_BuildAudioCVT (&wav_cvt,
                           wave.spec.format, wave.spec.channels,
                           wave.spec.freq, obtained->format,
                           obtained->channels, obtained->freq);

  /* Check that the convert was built */
  if (ret == -1)
    {
      fprintf (stderr, "Couldn't build converter!");
      SDL_FreeWAV (wave.sound);
      return 0;
    }

  /* Setup for conversion */
  wav_cvt.buf = (Uint8 *) malloc (wave.soundlen * wav_cvt.len_mult);
  wav_cvt.len = wave.soundlen;
  memcpy (wav_cvt.buf, wave.sound, wave.soundlen);

  /* We can delete to original WAV data now */
  SDL_FreeWAV (wave.sound);

  /* And now we're ready to convert */
  SDL_ConvertAudio (&wav_cvt);
  wave.sound = wav_cvt.buf;
  wave.soundlen *= wav_cvt.len_mult;

  return 1;
}

/* switch between the different musics */
int
switchmusic (void *dummy)
{
  (void) dummy;
  int i;
  int playnb;                   // number of the music which is going to be played
  static int played = 0;        // number of  music played
  static int startcycletime;
  static int done[MAXMUSIC];
  char *ptr;

  if (!startcycletime)
    startcycletime = time (NULL);

#    if (DEBUG)
  printf ("startcycletime %d\n", startcycletime);
  printf ("played %d\n", played);
#    endif
  if (nmusic == 0)
    return 1;
another:;
  if (played >= nmusic)
    {
#    if (DEBUG)
      printf ("new cycle \n");
#    endif
      for (i = 0; i < nmusic; i++)
        done[i] = 0;
      played = 0;
      if (time (NULL) - startcycletime <= 5)
        {
          printf ("Too short time in one cycle of the music\n");
          return 1;
        }
      startcycletime = time (NULL);
    }
  playnb = myrand () * nmusic;
  if (done[playnb])
    goto another;
#    if (DEBUG)
  printf ("playnb %d\n", playnb);
#    endif
  done[playnb] = 1;
  played++;
  ptr =
    malloc ((strlen (MUSICDIR "/") + strlen (files[playnb]) +
             1) * sizeof (char));
  sprintf (ptr, "%s%s", MUSICDIR "/", files[playnb]);
  if (!loadname (ptr))
    goto another;
  free (ptr);
#    if (DEBUG)
  printf ("playing %s\n", files[playnb]);
#    endif
  /* empty and fill the name of the music to show */
  free (music);
  music =
    malloc (sizeof (music) *
            (strstr (files[playnb], ".wav") - files[playnb] + 1));
  if (music != NULL)
    {
      strncpy (music, files[playnb],
               strstr (files[playnb], ".wav") - files[playnb]);
      music[strlen (files[playnb]) - 4] = '\0';
    }

  SDL_PauseAudio (0);
  if (ratio != 0.)
    forcedraw ();
  return 1;
}

/* filter to select possible music names */
int
musicfilter (const struct dirent *a)
{
  //if (strstr (a->d_name+(strlen(a->d_name)-4), ".wav")) return 1;
  const char *ext = getext (a->d_name);

  if (ext)
    {
      if (streq (ext, "wav"))
        return 1;
    }
  return 0;
}

/* initialize the sound playing process */
void
initializemusic (void)
{
  /* commented lines are obsolete : I used a system command to list the directory */
  //char * com = "dir -Nm *.wav";
  //char rep[MAXCMD] ;
  //char * pos;
  //int length=256;
  //int i = 0;
  //int nchar;
  struct dirent **namelist;
  int n;
  SDL_AudioSpec *desired;

  /* Load the SDL library */
  if (SDL_Init (SDL_INIT_AUDIO) < 0)
    {
      fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
      nosound = 1;
      playsound = 0;
      return;
    }
  atexit (SDL_Quit);

  /* prepare the audio converting */

  /* Allocated audio specs */
  desired = (SDL_AudioSpec *) malloc (sizeof (SDL_AudioSpec));
  obtained = (SDL_AudioSpec *) malloc (sizeof (SDL_AudioSpec));

  /* Set desired format */
  desired->freq = 44100;
  desired->format = AUDIO_S16LSB;
  desired->samples = 4096;
  desired->callback = fillerup;
  desired->channels = 2;
  desired->userdata = NULL;

  /* Open the audio device */
  if (SDL_OpenAudio (desired, obtained) < 0)
    {
      fprintf (stderr, "Couldn't open audio: %s", SDL_GetError ());
      nosound = 1;
      playsound = 0;
      SDL_Quit ();
      return;
    }

  free (desired);

/*psystem (com, strlen(com), rep, &length );
  next: pos = strstr (rep+i, ".wav");
        if (pos == NULL) goto done;
        nchar = pos-rep+4-i;
#if (DEBUG)
        printf ("nchar %d\n", nchar);
        printf ("music %d pos %d / %d\n", nmusic, i, length);
#endif
        strncpy (files[nmusic], rep+i, nchar);
        *(files[nmusic]+nchar) = '\0';
        i = i + nchar + 2 ;
#if (DEBUG)
        puts(files[nmusic]);
#endif
        nmusic++;
        if (nmusic >= MAXMUSIC) goto done;
        if (i >= length) goto done;
        goto next;
  done:;
#if (DEBUG)
  printf ("nmusic %d\n", nmusic );
#endif                                             */

  nmusic = myscandir (MUSICDIR, &namelist, musicfilter);
  if (nmusic == -1)
    {
      printf ("Problem listing the music directory "MUSICDIR
              " - you will probably experience problems...\n");
      printf ("Did you install the full data? (extra package available)\n");
      nmusic = 0;
      return;
    }
  if (nmusic > MAXMUSIC)
    nmusic = MAXMUSIC;
  for (n = 0; n < nmusic; n++)
    {
      strcpy (files[n], namelist[n]->d_name);
#    if (DEBUG)
      puts (files[n]);
#    endif
    }
#    if (DEBUG)
  printf ("nmusic %d\n", nmusic);
#    endif
}
#  endif
       // sdl





#  if (USESOUNDSDLMIXER)


/* load a music file */
int
loadname (char name[MAXCMD])
{
#    if (DEBUG)
  printf ("loading %s\n", name);
#    endif

  Mix_FreeChunk (wave.sound);
  /* load the song (wav, ogg, voc) */
  if (!(wave.sound = Mix_LoadWAV (name)))
    {
      printf ("error in Mix_LoadMUS(%s): %s\n", name, SDL_GetError ());
      return 0;
    }

  return 1;
}

/* switch between the musics */
int
switchmusic (void *dummy)
{
  (void) dummy;
  int i;
  int playnb;                   // number of the music which is going to be played
  static int played = 0;        // number of  music played
  static int startcycletime;
  static int done[MAXMUSIC];
  char *ptr;

  if (startcycletime == 0)
    startcycletime = time (NULL);

#    if (DEBUG)
  printf ("startcycletime %d\n", startcycletime);
  printf ("played %d\n", played);
#    endif
  if (nmusic == 0)
    return 1;
another:;
  if (played >= nmusic)
    {
#    if (DEBUG)
      printf ("new cycle \n");
#    endif
      for (i = 0; i < nmusic; i++)
        done[i] = 0;
      played = 0;
      if (time (NULL) - startcycletime <= 5)
        {
          printf ("Too short time in one cycle of the music\n");
          return 1;
        }
      startcycletime = time (NULL);
    }
  playnb = myrand () * nmusic;
  if (done[playnb])
    goto another;
#    if (DEBUG)
  printf ("playnb %d\n", playnb);
#    endif
  done[playnb] = 1;
  played++;
  ptr =
    malloc ((strlen (MUSICDIR "/") + strlen (files[playnb]) +
             1) * sizeof (char));
  sprintf (ptr, "%s%s", MUSICDIR "/", files[playnb]);
  if (!loadname (ptr))
    goto another;
  free (ptr);
#    if (DEBUG)
  printf ("playing %s\n", files[playnb]);
#    endif
  /* empty and fill the name of the music to show */
  free (music);
  music =
    malloc (sizeof (music) *
            (strstr (files[playnb], ".") - files[playnb] + 1));
  if (music != NULL)
    {
      strncpy (music, files[playnb],
               strstr (files[playnb], ".") - files[playnb]);
      music[strlen (files[playnb]) - 4] = '\0';
    }

  if (Mix_PlayChannel (0, wave.sound, 0) == -1)
    printf ("Mix_PlayChannel: %s\n", Mix_GetError ());

  if (ratio != 0.)
    forcedraw ();

  return 1;
}

/* SDL_Mixer called routine when a channel has finished playing */
void
endofchannel (int channel)
{
#    if (DEBUG)
  printf ("Channel %d finished\n", channel);
#    endif
  //if (channel == 0) switchmusic();
  if (channel == 0)
    SDL_CreateThread (switchmusic, NULL);
}

/* filter to select the right file's extensions */
int
musicfilter (const struct dirent *a)
{
  //if (strstr (a->d_name+(strlen(a->d_name)-4), ".ogg")) return 1;
  //if (strstr (a->d_name+(strlen(a->d_name)-4), ".wav")) return 1;
  //if (strstr (a->d_name+(strlen(a->d_name)-4), ".voc")) return 1;
#    if (DEBUG)
  printf ("checking: %s\n", a->d_name);
#    endif
  const char *ext = getext (a->d_name);

#    define CHECK(str) \
      if (streq (ext, str)) return 1

  if (ext)
    {
      CHECK ("ogg");
      CHECK ("wav");
      CHECK ("voc");
    }

#    undef CHECK

  return 0;
}

/* initialize the sound playing process */
void
initializemusic (void)
{
  struct dirent **namelist;
  int n;

  /* Load the SDL library */
  if (SDL_Init (SDL_INIT_AUDIO) < 0)
    {
      fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
      nosound = 1;
      playsound = 0;
      return;
    }
  atexit (SDL_Quit);

  /* Open the sdl_mixer audio device so loading on-the-fly is possible */
  if (Mix_OpenAudio (44100, MIX_DEFAULT_FORMAT, 2, 2048) < 0)
    {
      fprintf (stderr, "Couldn't open audio: %s\n", Mix_GetError ());
      nosound = 1;
      playsound = 0;
      SDL_Quit ();
      return;
    }
  nmusic = myscandir (MUSICDIR, &namelist, musicfilter);
  if (nmusic == -1)
    {
      printf ("Problem listing the music directory "MUSICDIR
              " - you will probably experience problems...\n");
      printf ("Did you install the full data? (extra package available)\n");
      nmusic = 0;
      return;
    }
  if (nmusic > MAXMUSIC)
    nmusic = MAXMUSIC;
  for (n = 0; n < nmusic; n++)
    {
      strcpy (files[n], namelist[n]->d_name);
#    if (DEBUG)
      puts (files[n]);
#    endif
    }
#    if (DEBUG)
  printf ("nmusic %d\n", nmusic);
#    endif

  Mix_AllocateChannels (LAST);
  for (n = SMENU1; n <= SMENU5; n++)
    Mix_Volume (n, MIX_MAX_VOLUME * (1 - (n - SMENU1) * 0.15));
  Mix_ChannelFinished (endofchannel);
}
#  endif
       /* sdl mixer */

#endif /* (USESOUNDSDL) || (USESOUNDSDLMIXER) */
